from collections.abc import Sequence
from dataclasses import dataclass

from xdsl.context import Context
from xdsl.dialects import arith, bufferization, func, linalg, memref, stencil, tensor
from xdsl.dialects.builtin import (
    AnyTensorType,
    AnyTensorTypeConstr,
    DenseArrayBase,
    DenseIntOrFPElementsAttr,
    FunctionType,
    MemRefType,
    ModuleOp,
    TensorType,
    i64,
)
from xdsl.dialects.csl import csl_stencil
from xdsl.ir import (
    Attribute,
    Block,
    BlockArgument,
    Operation,
    OpResult,
    Region,
    SSAValue,
)
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    TypeConversionPattern,
    attr_type_rewrite_pattern,
    op_type_rewrite_pattern,
)
from xdsl.rewriter import InsertPoint
from xdsl.utils.hints import isa
from xdsl.utils.isattr import isattr


def tensor_to_memref_type(t: TensorType[Attribute]) -> memref.MemRefType[Attribute]:
    """Type conversion from tensor to memref."""
    return memref.MemRefType(t.get_element_type(), t.get_shape())


def to_memref_op(op: SSAValue) -> bufferization.ToMemRefOp:
    """Creates a `bufferization.to_memref` operation."""
    assert isa(op.type, AnyTensorType)
    r_type = memref.MemRefType(
        op.type.get_element_type(), op.type.get_shape()
    )  # todo set strided+offset here?
    return bufferization.ToMemRefOp(operands=[op], result_types=[r_type])


def to_tensor_op(
    op: SSAValue, writable: bool = False, restrict: bool = True
) -> bufferization.ToTensorOp:
    """Creates a `bufferization.to_tensor` operation."""
    assert isa(op.type, MemRefType)
    return bufferization.ToTensorOp(op, restrict, writable)


class StencilTypeConversion(TypeConversionPattern):
    """
    Converts from tensorised stencil.field types to memref by extracting the element type which is a tensor
    and converting it to memref.

    For instance:
        `!stencil.field<[-1,1023]x[-1,511]xtensor<512xf32>` to `memref<512xf32>`
    """

    @attr_type_rewrite_pattern
    def convert_type(
        self, typ: stencil.FieldType[TensorType[Attribute]]
    ) -> memref.MemRefType[Attribute]:
        # todo should this convert to `memref` or `stencil.field<..xmemref<..>>`?
        return tensor_to_memref_type(typ.get_element_type())


@dataclass(frozen=True)
class ApplyOpBufferize(RewritePattern):
    """
    Bufferizes csl_stencil.apply, rewriting args and block args, changing them from tensor to memref types.
    For each converted arg, creates a `bufferization.to_memref` before the apply op.
    For each converted block arg, creates a `bufferization.to_tensor` at the start of the block.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: csl_stencil.ApplyOp, rewriter: PatternRewriter, /):
        if isa(op.accumulator.type, memref.MemRefType[Attribute]):
            return

        # convert args
        buf_args: list[SSAValue] = []
        to_memrefs: list[Operation] = [buf_iter_arg := to_memref_op(op.accumulator)]
        # in case of subsequent apply ops accessing this accumulator, replace uses with `bufferization.to_memref`
        op.accumulator.replace_by_if(
            buf_iter_arg.memref, lambda use: use.operation != buf_iter_arg
        )
        for arg in [*op.args_rchunk, *op.args_dexchng]:
            if isa(arg.type, TensorType[Attribute]):
                to_memrefs.append(new_arg := to_memref_op(arg))
                buf_args.append(new_arg.memref)
            else:
                buf_args.append(arg)

        # create new op
        buf_apply_op = csl_stencil.ApplyOp(
            operands=[
                op.field,
                buf_iter_arg.memref,
                op.args_rchunk,
                op.args_dexchng,
                op.dest,
            ],
            result_types=op.res.types or [[]],
            regions=[
                self._get_empty_bufferized_region(op.receive_chunk.block.args),
                self._get_empty_bufferized_region(op.done_exchange.block.args),
            ],
            properties=op.properties,
            attributes=op.attributes,
        )

        # insert to_tensor ops and create arg mappings for block inlining
        chunk_region_arg_mapping: Sequence[SSAValue] = []
        for idx, (old_arg, arg) in enumerate(
            zip(op.receive_chunk.block.args, buf_apply_op.receive_chunk.block.args)
        ):
            # arg0 has special meaning and does not need a `to_tensor` op
            if isinstance(old_arg.type, TensorType) and idx != 0:
                rewriter.insert_op(
                    # ensure iter_arg is writable
                    t := to_tensor_op(arg, writable=idx == 2),
                    InsertPoint.at_end(buf_apply_op.receive_chunk.block),
                )
                chunk_region_arg_mapping.append(t.tensor)
            else:
                chunk_region_arg_mapping.append(arg)

        done_exchange_arg_mapping: Sequence[SSAValue] = []
        for idx, (old_arg, arg) in enumerate(
            zip(op.done_exchange.block.args, buf_apply_op.done_exchange.block.args)
        ):
            if isinstance(old_arg.type, TensorType):
                rewriter.insert_op(
                    # ensure iter_arg is writable
                    t := to_tensor_op(arg, writable=idx == 1),
                    InsertPoint.at_end(buf_apply_op.done_exchange.block),
                )
                done_exchange_arg_mapping.append(t.tensor)
            else:
                done_exchange_arg_mapping.append(arg)

        assert isa(typ := op.receive_chunk.block.args[0].type, TensorType[Attribute])
        chunk_type = TensorType(typ.get_element_type(), typ.get_shape()[1:])

        # inline blocks from old into new regions
        rewriter.inline_block(
            op.receive_chunk.block,
            InsertPoint.at_end(buf_apply_op.receive_chunk.block),
            chunk_region_arg_mapping,
        )

        rewriter.inline_block(
            op.done_exchange.block,
            InsertPoint.at_end(buf_apply_op.done_exchange.block),
            done_exchange_arg_mapping,
        )

        self._inject_iter_arg_into_linalg_outs(
            buf_apply_op, rewriter, chunk_type, chunk_region_arg_mapping[2]
        )

        # insert new op
        rewriter.replace_matched_op(new_ops=[*to_memrefs, buf_apply_op])

    @staticmethod
    def _get_empty_bufferized_region(args: Sequence[BlockArgument]) -> Region:
        """Helper function to create a new region with bufferized arg types."""
        return Region(
            Block(
                arg_types=[
                    (
                        tensor_to_memref_type(arg.type)
                        if isattr(arg.type, AnyTensorTypeConstr)
                        else arg.type
                    )
                    for arg in args
                ]
            )
        )

    @staticmethod
    def _inject_iter_arg_into_linalg_outs(
        op: csl_stencil.ApplyOp,
        rewriter: PatternRewriter,
        chunk_type: TensorType[Attribute],
        iter_arg: SSAValue,
    ):
        """
        Finds a linalg op with `chunk_type` shape in `outs` and injects
        an extracted slice of `iter_arg`. This is a work-around for the
        way bufferization works, causing it to use `iter_arg` as an accumulator
        and avoiding having an extra alloc + memref.copy.
        """
        linalg_op: linalg.NamedOpBase | None = None
        for curr_op in op.receive_chunk.block.ops:
            if (
                isinstance(curr_op, linalg.NamedOpBase)
                and len(curr_op.outputs) > 0
                and curr_op.outputs.types[0] == chunk_type
            ):
                linalg_op = curr_op
                break

        if linalg_op is None:
            return

        rewriter.replace_op(
            linalg_op,
            [
                extract_slice_op := tensor.ExtractSliceOp(
                    operands=[iter_arg, [op.receive_chunk.block.args[1]], [], []],
                    result_types=[chunk_type],
                    properties={
                        "static_offsets": DenseArrayBase.from_list(
                            i64, (memref.SubviewOp.DYNAMIC_INDEX,)
                        ),
                        "static_sizes": DenseArrayBase.from_list(
                            i64, chunk_type.get_shape()
                        ),
                        "static_strides": DenseArrayBase.from_list(i64, (1,)),
                    },
                ),
                type(linalg_op).build(
                    operands=[linalg_op.inputs, extract_slice_op.results],
                    result_types=linalg_op.result_types,
                    properties=linalg_op.properties,
                    attributes=linalg_op.attributes,
                    regions=[linalg_op.detach_region(r) for r in linalg_op.regions],
                ),
            ],
        )

    @staticmethod
    def _build_extract_slice(
        op: csl_stencil.ApplyOp, to_tensor: bufferization.ToTensorOp, offset: SSAValue
    ) -> tensor.ExtractSliceOp:
        """
        Helper function to create an early tensor.extract_slice in the apply.recv_chunk_cb region needed for bufferization.
        """

        # this is the unbufferized `tensor<(neighbours)x(ZDim)x(type)>` value
        assert isa(typ := op.receive_chunk.block.args[0].type, TensorType[Attribute])

        return tensor.ExtractSliceOp(
            operands=[to_tensor.tensor, [offset], [], []],
            result_types=[TensorType(typ.get_element_type(), typ.get_shape()[1:])],
            properties={
                "static_offsets": DenseArrayBase.from_list(
                    i64, (memref.SubviewOp.DYNAMIC_INDEX,)
                ),
                "static_sizes": DenseArrayBase.from_list(i64, typ.get_shape()[1:]),
                "static_strides": DenseArrayBase.from_list(i64, (1,)),
            },
        )


@dataclass(frozen=True)
class AccessOpBufferize(RewritePattern):
    """
    Bufferizes AccessOp.

    The type conversion pass creates the scenario that some `csl_stencil.access` ops are equal input and output types,
    for instance, `(memref<512xf32>) -> memref<512xf32>`. This only happens for ops accessing own data. In this case,
    the access op has no effect and can safely be folded away.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: csl_stencil.AccessOp, rewriter: PatternRewriter, /):
        if not isa(op.result.type, TensorType[Attribute]):
            return
        r_type = tensor_to_memref_type(op.result.type)

        # accesses to own data that (after bufferization) have the same input and output type can be safely folded away
        if op.op.type == r_type and all(o == 0 for o in op.offset):
            rewriter.replace_matched_op(to_tensor_op(op.op))
            return

        # accesses to buffers passed in additional args can read directly from memref underlying `to_tensor`
        source = (
            op.op.op.memref
            if isinstance(op.op, OpResult)
            and isinstance(op.op.op, bufferization.ToTensorOp)
            else op.op
        )

        rewriter.replace_matched_op(
            [
                access := csl_stencil.AccessOp(
                    source,
                    op.offset,
                    r_type,
                    op.offset_mapping,
                ),
                to_tensor_op(access.result),
            ]
        )


@dataclass(frozen=True)
class YieldOpBufferize(RewritePattern):
    """Bufferizes YieldOp."""

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: csl_stencil.YieldOp, rewriter: PatternRewriter, /):
        to_memrefs: list[Operation] = []
        args: list[SSAValue] = []
        for arg in op.arguments:
            if isa(arg.type, TensorType[Attribute]):
                to_memrefs.append(new_arg := to_memref_op(arg))
                args.append(new_arg.memref)
            else:
                args.append(arg)

        if len(to_memrefs) == 0:
            return

        rewriter.replace_matched_op([*to_memrefs, csl_stencil.YieldOp(*args)])


@dataclass(frozen=True)
class FuncOpBufferize(RewritePattern):
    """
    Replace the function_type and let a separate type conversion pass handle the block args.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: func.FuncOp, rewriter: PatternRewriter, /):
        function_type = FunctionType.from_lists(
            [
                (
                    tensor_to_memref_type(t.get_element_type())
                    if isa(t, stencil.FieldType[TensorType[Attribute]])
                    else t
                )
                for t in op.function_type.inputs
            ],
            [
                (
                    tensor_to_memref_type(t.get_element_type())
                    if isa(t, stencil.FieldType[TensorType[Attribute]])
                    else t
                )
                for t in op.function_type.outputs
            ],
        )
        if function_type == op.function_type:
            return
        rewriter.replace_matched_op(
            func.FuncOp.build(
                operands=op.operands,
                result_types=op.result_types,
                regions=[op.detach_region(op.body)],
                properties={**op.properties, "function_type": function_type},
                attributes=op.attributes.copy(),
            )
        )


@dataclass(frozen=True)
class ArithConstBufferize(RewritePattern):
    """
    Bufferize arith tensor constants to prevent mlir bufferize from promoting them to globals.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.ConstantOp, rewriter: PatternRewriter, /):
        if not isa(op.result.type, TensorType[Attribute]):
            return
        assert isinstance(op.value, DenseIntOrFPElementsAttr)
        assert isa(op.value.type, TensorType[Attribute])
        typ = DenseIntOrFPElementsAttr(
            [tensor_to_memref_type(op.value.type), op.value.data]
        )
        rewriter.replace_matched_op(
            [
                c := arith.ConstantOp(typ),
                to_tensor_op(c.result),
            ]
        )


@dataclass(frozen=True)
class InjectApplyOutsIntoLinalgOuts(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: csl_stencil.ApplyOp, rewriter: PatternRewriter, /):
        # require bufferized apply (with op.dest specified)
        # zero-output apply ops may be used for communicate-only, to which this pattern does not apply
        if not op.dest:
            return

        yld = op.done_exchange.block.last_op
        assert isinstance(yld, csl_stencil.YieldOp)
        new_dest: list[SSAValue] = []
        new_yield_args: list[SSAValue] = []
        additional_args: list[SSAValue] = []
        to_remove: list[Operation] = []

        for arg, yld_arg in zip(op.dest, yld.arguments, strict=True):
            if (
                not isinstance(yld_arg, OpResult)
                or not isinstance(yld_arg.op, bufferization.ToMemRefOp)
                or not isinstance(yld_arg.op.tensor, OpResult)
                or not isinstance(
                    linalg_op := yld_arg.op.tensor.op,
                    linalg.NamedOpBase | linalg.GenericOp,
                )
                or not isa(arg_t := arg.type, MemRefType)
                or not isa(yld_arg.type, MemRefType)
            ):
                new_dest.append(arg)
                new_yield_args.append(yld_arg)
                continue
            additional_args.append(arg)
            if len(yld_arg.uses) == 1:
                to_remove.append(yld_arg.op)

            arg = op.done_exchange.block.insert_arg(
                arg.type, len(op.done_exchange.block.args)
            )
            arg_to_tensor = to_tensor_op(arg, writable=True)

            # offset of core data, assuming symmetric ghost cells in each direction
            offsets = tuple(
                (src - dst) // 2  # symmetric offset
                for src, dst in zip(
                    arg_t.get_shape(), yld_arg.type.get_shape(), strict=True
                )
            )

            extract_slice_op = tensor.ExtractSliceOp(
                operands=[arg_to_tensor, [], [], []],
                result_types=[yld_arg.op.tensor.type],
                properties={
                    "static_offsets": DenseArrayBase.from_list(i64, offsets),
                    "static_sizes": DenseArrayBase.from_list(
                        i64, yld_arg.type.get_shape()
                    ),
                    "static_strides": DenseArrayBase.from_list(i64, (1,)),
                },
            )
            rewriter.insert_op(
                [arg_to_tensor, extract_slice_op], InsertPoint.before(linalg_op)
            )
            rewriter.replace_op(
                linalg_op,
                type(linalg_op).build(
                    operands=[linalg_op.inputs, [extract_slice_op.result]],
                    result_types=linalg_op.result_types,
                    regions=[linalg_op.detach_region(r) for r in linalg_op.regions],
                    properties=linalg_op.properties,
                    attributes=linalg_op.attributes,
                ),
            )
        if additional_args:
            rewriter.replace_op(yld, csl_stencil.YieldOp(*new_yield_args))
            for r in to_remove:
                rewriter.erase_op(r)
            rewriter.replace_matched_op(
                csl_stencil.ApplyOp(
                    operands=[
                        op.field,
                        op.accumulator,
                        [*op.args_rchunk],
                        [*op.args_dexchng, *additional_args],
                        [*op.dest],
                    ],
                    result_types=op.res.types or [[]],
                    regions=[op.detach_region(r) for r in op.regions],
                    properties=op.properties,
                    attributes=op.attributes,
                )
            )


@dataclass(frozen=True)
class ReselectLinalgOutsFromInputs(RewritePattern):
    """
    Reselects the DPS `outs` of a linalg op if it is one of its inputs.
      * select `writable` tensor input with no later use, or else
      * select linalg op input with no later use
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: linalg.NamedOpBase | linalg.GenericOp, rewriter: PatternRewriter, /
    ):
        # only apply rewrite when re-selecting `outs` from `ins`
        if (
            op.outputs[0] not in op.inputs
            or self.is_writable(op.outputs[0])
            or len(op.outputs) != 1
        ):
            return

        # the new `outs` to re-select
        out: SSAValue | None = None

        for arg in op.inputs:
            # reselect outs that has no later use to avoid read-after-write conflicts
            if len(arg.uses) == 1:
                # check for a `writable` input with no later uses and break immediately
                if self.is_writable(arg):
                    out = arg
                    break

                # check for a linalg op input with no later uses and keep looking
                if isinstance(arg, OpResult) and isinstance(
                    arg.op, linalg.NamedOpBase | linalg.GenericOp
                ):
                    out = arg

        # replace the op with `out` as `output[0]`
        if out:
            rewriter.replace_matched_op(
                type(op).build(
                    operands=[op.inputs, [out]],
                    result_types=op.result_types,
                    regions=[op.detach_region(r) for r in op.regions],
                    properties=op.properties,
                    attributes=op.attributes,
                )
            )

    @staticmethod
    def is_writable(val: SSAValue) -> bool:
        """Returns if `val` is a `writable` tensor."""
        return (
            isinstance(val, OpResult)
            and isinstance(val.op, bufferization.ToTensorOp)
            and val.op.writable is not None
        )


@dataclass(frozen=True)
class CslStencilBufferize(ModulePass):
    """
    Bufferizes the csl_stencil dialect.

    Attempts to inject `csl_stencil.apply.recv_chunk_cb.accumulator` into linalg compute ops `outs` within that region
    for improved bufferization. Ideally be run after `--lift-arith-to-linalg`.

    In preparation for bufferization with minimal overhead, linalg ops `outs` are set as follows:
      if a linalg op's destination is one of its inputs
      1. prefer a `writable` input with no other uses
      2. prefer a linalg op input with no other uses
    """

    name = "csl-stencil-bufferize"

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        module_pass = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    StencilTypeConversion(),
                    ApplyOpBufferize(),
                    AccessOpBufferize(),
                    YieldOpBufferize(),
                    FuncOpBufferize(),
                    ArithConstBufferize(),
                ]
            )
        )
        module_pass.rewrite_module(op)
        PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    InjectApplyOutsIntoLinalgOuts(),
                    ReselectLinalgOutsFromInputs(),
                ]
            ),
            apply_recursively=False,
        ).rewrite_module(op)



class LazyVersion:
    """
    Resolving the version dynamically is slow, hence this lazy wrapper to get the version
    only when needed.
    """

    def __init__(self):
        self._version = None

    def __str__(self):
        if self._version is None:
            from . import _version

            self._version = _version.get_versions()["version"]
        return self._version


__version__ = LazyVersion()



"""
The High-Level Fortran IR (HLFIR) dialect that is used by Flang.

This is mixed with FIR, and provides a higher level view of Fortran
variables and some expressions. This means that temporaries need not
be materialised and there is a higher level of information about
the programmer's Fortran code, compared to FIR, available for
optimisation and lowering.

See external [documentation](https://flang.llvm.org/docs/HighLevelFIR.html).
"""

from __future__ import annotations

from xdsl.dialects.arith import FastMathFlagsAttr
from xdsl.dialects.builtin import (
    AnyFloat,
    ArrayAttr,
    Attribute,
    BoolAttr,
    DenseArrayBase,
    IntAttr,
    IntegerAttr,
    IntegerType,
    ParametrizedAttribute,
    StringAttr,
    UnitAttr,
)
from xdsl.dialects.experimental.fir import (
    DeferredAttr,
    FortranVariableFlagsAttr,
    NoneType,
    ReferenceType,
)
from xdsl.ir import Dialect, TypeAttribute
from xdsl.irdl import (
    AttrSizedOperandSegments,
    IRDLOperation,
    ParameterDef,
    irdl_attr_definition,
    irdl_op_definition,
    operand_def,
    opt_operand_def,
    opt_prop_def,
    prop_def,
    region_def,
    result_def,
    var_operand_def,
)
from xdsl.parser import AttrParser
from xdsl.printer import Printer


@irdl_attr_definition
class ExprType(ParametrizedAttribute, TypeAttribute):
    """
    Abstract value type for Fortran arrays, characters and derived types.
    The rank cannot be assumed, and empty shape means that the expression is a scalar.
    When the element type is a derived type, the polymorphic flag may be set to true
    to indicate that the expression dynamic type can differ from its static type.
    """

    name = "hlfir.expr"
    shape: ParameterDef[ArrayAttr[IntegerAttr | DeferredAttr | NoneType]]
    elementType: ParameterDef[IntegerType | AnyFloat | ReferenceType]

    def print_parameters(self, printer: Printer) -> None:
        printer.print("<")
        for s in self.shape.data:
            if isinstance(s, DeferredAttr):
                printer.print_string("?")
            elif isinstance(s, NoneType):
                raise Exception(
                    "Can not have none type as part of sequence shape with only one type"
                )
            else:
                printer.print_string(f"{s.value.data}")
            printer.print_string("x")
        printer.print(self.elementType)

        printer.print(">")

    @classmethod
    def parse_parameters(cls, parser: AttrParser) -> list[Attribute]:
        # We need extra work here as the builtin tuple is not being supported
        # yet, therefore handle this here
        def parse_interval() -> IntegerAttr[IntegerType] | DeferredAttr:
            if parser.parse_optional_punctuation("?"):
                return DeferredAttr()
            s = parser.parse_integer(allow_boolean=False)
            return IntegerAttr[IntegerType](s, 32)

        shape: list[IntegerAttr[IntegerType] | DeferredAttr] = []
        parser.parse_characters("<")
        elementType = parser.parse_optional_type()
        while elementType is None:
            shape.append(parse_interval())
            parser.parse_shape_delimiter()
            elementType = parser.parse_optional_type()
        parser.parse_characters(">")
        return [ArrayAttr(shape), elementType]


@irdl_op_definition
class DeclareOp(IRDLOperation):
    """
    Tie the properties of a Fortran variable to an address. The properties
    include bounds, length parameters, and Fortran attributes.

    The arguments are the same as for fir.declare.

    The main difference with fir.declare is that hlfir.declare returns two
    values:
      - the first one is an SSA value that allows retrieving the variable
        address, bounds, and type parameters at any point without requiring
        access to the defining operation. This may be:
        - for scalar numerical, logical, or derived type without length
          parameters: a fir.ref<T> (e.g. fir.ref<i32>)
        - for scalar characters: a fir.boxchar<kind> or fir.ref<fir.char<kind,
          cst_len>>
        - for arrays of types without length parameters, without lower bounds,
          that are not polymorphic and with a constant shape:
          fir.ref<fir.array<cst_shapexT>>
        - for all non pointer/non allocatable entities: fir.box<T>, and
          fir.class<T> for polymorphic entities.
        - for all pointers/allocatables:
          fir.ref<fir.box<fir.ptr<T>>>/fir.ref<fir.box<fir.heap<T>>>
      - the second value has the same type as the input memref, and is the
        same. If it is a fir.box or fir.class, it may not contain accurate
        local lower bound values. It is intended to be used when generating FIR
        from HLFIR in order to avoid descriptor creation for simple entities.

    Example:
    CHARACTER(n) :: c(10:n, 20:n)

    Can be represented as:
    func.func @foo(%arg0: !fir.ref<!fir.array<?x?x!fir.char<1,?>>>, %arg1: !fir.ref<i64>) {
      %c10 = arith.constant 10 : index
      %c20 = arith.constant 20 : index
      %1 = fir.load %ag1 : fir.ref<i64>
      %2 = fir.shape_shift %c10, %1, %c20, %1 : (index, index, index, index) -> !fir.shapeshift<2>
      %3 = hfir.declare %arg0(%2) typeparams %1 {uniq_name = "c"} (fir.ref<!fir.array<?x?x!fir.char<1,?>>>, fir.shapeshift<2>, index) -> (fir.box<!fir.array<?x?x!fir.char<1,?>>>, fir.ref<!fir.array<?x?x!fir.char<1,?>>>)
      // ... uses %3#0 as "c"
    }
    """  # noqa: E501

    name = "hlfir.declare"
    memref = operand_def()
    shape = opt_operand_def()
    typeparams = var_operand_def()
    uniq_name = opt_prop_def(StringAttr)
    fortran_attrs = opt_prop_def(FortranVariableFlagsAttr)
    result = result_def()
    result2 = result_def()

    irdl_options = [AttrSizedOperandSegments(as_property=True)]


@irdl_op_definition
class DesignateOp(IRDLOperation):
    """
    This operation represents a Fortran "part-ref", except that it can embed a
    substring or or complex part directly, and that vector subscripts cannot be
    used. It returns a Fortran variable that is a part of the input variable.

    The operands are as follow:
      - memref is the variable being designated.
      - component may be provided if the memref is a derived type to
        represent a reference to a component. It must be the name of a
        component of memref derived type.
      - component_shape represents the shape of the component and must be
        provided if and only if both component and indices appear.
      - indices can be provided to index arrays. The indices may be simple
        indices or triplets.
        If indices are provided and there is a component, the component must be
        an array component and the indices index the array component.
        If memref is an array, and component is provided and is an array
        component, indices must be provided and must not be triplets. This
        ensures hlfir.designate does not create arrays of arrays (which is not
        possible in Fortran).
      - substring may contain two values to represent a substring lower and
        upper bounds.
      - complex_part may be provided to represent a complex part (true
        represents the imaginary part, and false the real part).
      - shape represents the shape of the result and must be provided if the
        result is an array that is not a box address.
      - typeparams represents the length parameters of the result and must be
        provided if the result type has length parameters and is not a box
        address.
    """

    name = "hlfir.designate"
    memref = operand_def()
    component = opt_prop_def(StringAttr)
    component_shape = opt_operand_def()
    indices = var_operand_def()
    substring = var_operand_def()
    complex_part = opt_prop_def(BoolAttr)
    shape = opt_operand_def()
    typeparams = var_operand_def()
    is_triplet = prop_def(DenseArrayBase)
    fortran_attrs = opt_prop_def(FortranVariableFlagsAttr)
    result = result_def()

    irdl_options = [AttrSizedOperandSegments(as_property=True)]


@irdl_op_definition
class AssignOp(IRDLOperation):
    """
    Assign rhs to lhs following Fortran intrinsic assignments rules.
    The operation deals with inserting a temporary if the lhs and rhs
    may overlap.

    The optional "realloc" flag allows indicating that this assignment
    has the Fortran 95 semantics for assignments to a whole allocatable.
    In such case, the left hand side must be an allocatable that may be
    unallocated or allocated with a different type and shape than the right
    hand side. It will be allocated or re-allocated as needed during the
    assignment.

    When "realloc" is set and this is a character assignment, the optional
    flag "keep_lhs_length_if_realloc" indicates that the character
    left hand side should retain its length after the assignment. If the
    right hand side has a different length, truncation and padding will
    occur. This covers the case of explicit and assumed length character
    allocatables.

    Otherwise, the left hand side will be allocated or reallocated to match the
    right hand side length if they differ. This covers the case of deferred
    length character allocatables.

    The optional "temporary_lhs" flag indicates that the LHS is a compiler
    generated temporary. In this case the temporary is initialized if needed
    (e.g. the LHS is of derived type with allocatable/pointer components),
    and the assignment is done without LHS (or its subobjects) finalization
    and with automatic allocation.

    If "temporary_lhs" and "keep_lhs_length_if_realloc" are both set,
    this assign operation denotes special case of character allocatable
    LHS with explicit length. The LHS that must preserve its length
    during the assignment regardless of the the RHS's length or/and
    allocation status. This assign operation will be lowered into a call
    to AssignExplicitLengthCharacter().
    """

    name = "hlfir.assign"
    lhs = operand_def()
    rhs = operand_def()
    realloc = opt_prop_def(UnitAttr)
    keep_lhs_length_if_realloc = opt_prop_def(UnitAttr)
    temporary_lhs = opt_prop_def(UnitAttr)


@irdl_op_definition
class ParentComponentOp(IRDLOperation):
    """
    This operation represents a Fortran component reference where the
    component name is a parent type of the variable's derived type.
    These component references cannot be represented with an hlfir.designate
    because the parent type names are not embedded in fir.type<> types
    as opposed to the actual component names.

    The operands are as follow:
      - memref is a derived type variable whose parent component is being
        designated.
      - shape is the shape of memref and the result and must be provided if
        memref is an array. Parent component reference lower bounds are ones,
        so the provided shape must be a fir.shape.
      - typeparams are the type parameters of the parent component type if any.
        It is a subset of memref type parameters.
    The parent component type and name is reflected in the result type.
    """

    name = "hlfir.parent_comp"
    memref = operand_def()
    shape = opt_operand_def()
    typeparams = var_operand_def()
    result = result_def()

    irdl_options = [AttrSizedOperandSegments(as_property=True)]


@irdl_op_definition
class ConcatOp(IRDLOperation):
    """
    Concatenate two or more character strings of a same character kind.
    """

    name = "hlfir.concat"
    strings = var_operand_def()
    length = operand_def()

    irdl_options = [AttrSizedOperandSegments(as_property=True)]


@irdl_op_definition
class AllOp(IRDLOperation):
    """
    Takes a logical array MASK as argument, optionally along a particular dimension,
    and returns true if all elements of MASK are true.
    """

    name = "hlfir.all"
    mask = operand_def()
    dim = opt_operand_def()
    result = result_def()


@irdl_op_definition
class AnyOp(IRDLOperation):
    """
    Takes a logical array MASK as argument, optionally along a particular dimension,
    and returns true if any element of MASK is true.
    """

    name = "hlfir.any"
    mask = operand_def()
    dim = opt_operand_def()
    result = result_def()


@irdl_op_definition
class CountOp(IRDLOperation):
    """
    Takes a logical and counts the number of true values.
    """

    name = "hlfir.count"
    mask = operand_def()
    dim = opt_operand_def()
    kind = opt_operand_def()
    result = result_def()

    irdl_options = [AttrSizedOperandSegments(as_property=True)]


@irdl_op_definition
class MaxvalOp(IRDLOperation):
    """
    Maximum value(s) of an array.
    If DIM is absent, the result is a scalar.
    If DIM is present, the result is an array of rank n-1, where n is the rank of ARRAY.
    """

    name = "hlfir.maxval"
    array = operand_def()
    mask = opt_operand_def()
    dim = opt_operand_def()
    fastmath = opt_prop_def(FastMathFlagsAttr)
    result = result_def()

    irdl_options = [AttrSizedOperandSegments(as_property=True)]


@irdl_op_definition
class MinvalOp(IRDLOperation):
    """
    Minimum value(s) of an array.
    If DIM is absent, the result is a scalar.
    If DIM is present, the result is an array of rank n-1, where n is the rank of ARRAY.
    """

    name = "hlfir.minval"
    array = operand_def()
    mask = opt_operand_def()
    dim = opt_operand_def()
    fastmath = opt_prop_def(FastMathFlagsAttr)
    result = result_def()

    irdl_options = [AttrSizedOperandSegments(as_property=True)]


@irdl_op_definition
class ProductOp(IRDLOperation):
    """
    Multiplies the elements of an array, optionally along a particular dimension,
    optionally if a mask is true.
    """

    name = "hlfir.product"
    array = operand_def()
    mask = opt_operand_def()
    dim = opt_operand_def()
    fastmath = opt_prop_def(FastMathFlagsAttr)
    result = result_def()

    irdl_options = [AttrSizedOperandSegments(as_property=True)]


@irdl_op_definition
class SetLengthOp(IRDLOperation):
    """
    Change the length of character entity. This trims or pads the
    character argument according to the new length.
    """

    name = "hlfir.set_length"
    string = operand_def()
    length = operand_def()
    result = result_def()


@irdl_op_definition
class GetLengthOp(IRDLOperation):
    """
    Get the length of character entity represented as hlfir.expr.
    """

    name = "hlfir.get_length"
    expr = operand_def()
    result = result_def()


@irdl_op_definition
class SumOp(IRDLOperation):
    """
    Sums the elements of an array, optionally along a particular dimension,
    optionally if a mask is true.
    """

    name = "hlfir.sum"
    array = operand_def()
    mask = opt_operand_def()
    dim = opt_operand_def()
    fastmath = opt_prop_def(FastMathFlagsAttr)
    result = result_def()

    irdl_options = [AttrSizedOperandSegments(as_property=True)]


@irdl_op_definition
class DotProductOp(IRDLOperation):
    """
    Dot product of two vectors
    """

    name = "hlfir.dot_product"
    lhs = operand_def()
    rhs = operand_def()
    fastmath = opt_prop_def(FastMathFlagsAttr)
    result = result_def()


@irdl_op_definition
class MatmulOp(IRDLOperation):
    """
    Matrix multiplication
    """

    name = "hlfir.matmul"
    lhs = operand_def()
    rhs = operand_def()
    fastmath = opt_prop_def(FastMathFlagsAttr)
    result = result_def()


@irdl_op_definition
class TransposeOp(IRDLOperation):
    """
    Transpose a rank 2 array
    """

    name = "hlfir.transpose"
    array = operand_def()
    result = result_def()


@irdl_op_definition
class MatmulTransposeOp(IRDLOperation):
    """
    Matrix multiplication where the left hand side is transposed
    """

    name = "hlfir.matmul_transpose"
    lhs = operand_def()
    rhs = operand_def()
    fastmath = opt_prop_def(FastMathFlagsAttr)
    result = result_def()


@irdl_op_definition
class AssociateOp(IRDLOperation):
    """
    Create a variable from an expression value.
    For expressions, this operation is an incentive to re-use the expression
    storage, if any, after the bufferization pass when possible (if the
    expression is not used afterwards).
    """

    name = "hlfir.associate"
    source = operand_def()
    shape = opt_operand_def()
    typeparams = var_operand_def()
    uniq_name = opt_prop_def(StringAttr)
    fortran_attrs = opt_prop_def(FortranVariableFlagsAttr)
    result = result_def()

    irdl_options = [AttrSizedOperandSegments(as_property=True)]


@irdl_op_definition
class EndAssociateOp(IRDLOperation):
    """
    Mark the end of life of a variable associated to an expression.
    If the expression has a derived type that may contain allocatable
    components, the variable operand must be a Fortran entity.
    """

    name = "hlfir.end_associate"
    var = operand_def()
    must_free = operand_def()


@irdl_op_definition
class AsExprOp(IRDLOperation):
    """
    Take the value of an array, character or derived variable.
    In general, this operation will lead to a copy of the variable
    in the bufferization pass if it was not transformed.

    However, if it is known that the variable storage will not be used anymore
    afterwards, the variable storage ownership can be passed to the hlfir.expr
    by providing the $must_free argument that is a boolean that indicates if
    the storage must be freed (when it was allocated on the heap).
    This allows Fortran lowering to build some expression value in memory when
    there is no adequate hlfir operation, and to promote the result to an
    hlfir.expr value without paying the price of introducing a copy.
    """

    name = "hlfir.as_expr"
    var = operand_def()
    must_free = opt_operand_def()
    result = result_def()


@irdl_op_definition
class NoReassocOp(IRDLOperation):
    """
    Same as fir.reassoc, except it accepts hlfir.expr arguments.
    """

    name = "hlfir.no_reassoc"
    var = operand_def()
    result = result_def()


@irdl_op_definition
class ElementalOp(IRDLOperation):
    """
    Represent an elemental expression as a function of the indices.
    This operation contain a region whose block arguments are one
    based indices iterating over the elemental expression shape.
    Given these indices, the element value for the given iteration
    can be computed in the region and yielded with the hlfir.yield_element
    operation.

    The shape and typeparams operands represent the extents and type
    parameters of the resulting array value.

    The optional mold is an entity carrying the information about
    the dynamic type of the polymorphic result. Note that the shape
    of the mold does not necessarily match the shape of the result,
    for example, the result of `merge(poly_scalar1, poly_scalar2, mask_array)`
    will have the shape of `mask_array` and the dynamic type of `poly_scalar*`.

    The unordered attribute can be set to allow out of order processing
    of the indices. This is safe only if the operations in the body
    of the elemental do not have side effects.


    Example: Y + X,  with Integer :: X(10, 20), Y(10,20)

    %0 = fir.shape %c10, %c20 : (index, index) -> !fir.shape<2>
    %5 = hlfir.elemental %0 : (!fir.shape<2>) -> !hlfir.expr<10x20xi32> {
    ^bb0(%i: index, %j: index):
      %6 = hlfir.designate %x (%i, %j)  : (!fir.ref<!fir.array<10x20xi32>>, index, index) -> !fir.ref<i32>
      %7 = hlfir.designate %y (%i, %j)  : (!fir.ref<!fir.array<10x20xi32>>, index, index) -> !fir.ref<i32>
      %8 = fir.load %6 : !fir.ref<i32>
      %9 = fir.load %7 : !fir.ref<i32>
      %10 = arith.addi %8, %9 : i32
      hlfir.yield_element %10 : i32
    }
    """

    name = "hlfir.elemental"
    shape = operand_def()
    mold = opt_operand_def()
    typeparams = var_operand_def()
    unordered = opt_prop_def(UnitAttr)
    result = result_def()

    irdl_options = [AttrSizedOperandSegments(as_property=True)]


@irdl_op_definition
class YieldElementOp(IRDLOperation):
    """
    Yield the element value of the current elemental expression iteration
    in an hlfir.elemental region. See hlfir.elemental description for an
    example.
    """

    name = "hlfir.yield_element"
    element_value = operand_def()


@irdl_op_definition
class ApplyOp(IRDLOperation):
    """
    Given an hlfir.expr array value, hlfir.apply allow retrieving
    the value for an element given one based indices.

    When hlfir.apply is used on an hlfir.elemental, and if the hlfir.elemental
    operation evaluation can be moved to the location of the hlfir.apply, it is
    as if the hlfir.elemental body was evaluated given the hlfir.apply indices.
    Therefore, apply operations on hlfir.elemental expressions should be located
    such that evaluating the hlfir.elemental at the position of the hlfir.apply
    operation produces the same result as evaluating the hlfir.elemental at its
    location in the instruction stream. Attention should be paid to
    hlfir.elemental memory side effects (in practice these are unlikely).
    "10.1.4 Evaluation of operations" says that expression evaluation shall not
    impact/be impacted by other expression evaluation in the statement.
    """

    name = "hlfir.apply"
    expr = operand_def()
    indices = var_operand_def()
    typeparams = var_operand_def()
    result = result_def()

    irdl_options = [AttrSizedOperandSegments(as_property=True)]


@irdl_op_definition
class NullOp(IRDLOperation):
    """
    Create a NULL() address.
    So far is not intended to represent NULL(MOLD).
    """

    name = "hlfir.null"
    result = result_def()


@irdl_op_definition
class DestroyOp(IRDLOperation):
    """
    Mark the last use of an hlfir.expr. This will be the point at which the
    buffer of an hlfir.expr, if any, will be deallocated if it was heap
    allocated.
    If "finalize" attribute is set, the hlfir.expr value will be finalized
    before the deallocation. Note that this implies that the hlfir.expr
    is placed into a memory buffer, so that the library runtime
    can be called on it. The element type of the hlfir.expr must be
    derived type in this case.
    It is not required to create an hlfir.destroy operation for and hlfir.expr
    created inside an hlfir.elemental and returned in the hlfir.yield_element.
    The last use of such expression is implicit and an hlfir.destroy could
    not be emitted after the hlfir.yield_element since it is a terminator.

    Note that hlfir.destroy are currently generated by Fortran lowering that
    has a good view of the expression use contexts, but this will need to be
    revisited if any motion of hlfir.expr is done (like CSE) since
    transformations should not introduce any hlfir.expr usages after an
    hlfir.destroy.
    The future will probably be to identify the last use points automatically
    in bufferization instead.
    """

    name = "hlfir.destroy"
    expr = operand_def()
    finalize = opt_prop_def(UnitAttr)


@irdl_op_definition
class CopyInOp(IRDLOperation):
    """
    Copy a variable into a contiguous temporary if the variable is not
    an absent optional and is not contiguous at runtime. When a copy is made this
    operation returns the temporary as first result, otherwise, it returns the
    potentially absent variable storage. The second result indicates if a copy
    was made.

    This operation is meant to be used in combination with the hlfir.copy_out
    operation that deletes the temporary if it was created and copies the data
    back if needed.
    This operation allows passing non contiguous arrays to contiguous dummy
    arguments, which is possible in Fortran procedure references.

    To deal with the optional case, an extra boolean value can be pass to the
    operation. In such cases, the copy-in will only be done if "var_is_present"
    is true and, when it is false, the original value will be returned instead.
    """

    name = "hlfir.copy_in"
    var = operand_def()
    var_is_present = opt_operand_def()
    result = result_def()


@irdl_op_definition
class CopyOutOp(IRDLOperation):
    """
    If the variable was copied in a temporary in the related hlfir.copy_in,
    optionally copy back the temporary value to it (that may have been
    modified between the hlfir.copy_in and hlfir.copy_out). Then deallocate
    the temporary.
    The copy back is done if $var is provided and $was_copied is true.
    The deallocation of $temp is done if $was_copied is true.
    """

    name = "hlfir.copy_out"
    temp = operand_def()
    was_copied = operand_def()
    var = opt_operand_def()


@irdl_op_definition
class ShapeOfOp(IRDLOperation):
    """
    Gets the runtime shape of a hlfir.expr. In lowering to FIR, the
    hlfir.shape_of operation will be replaced by an fir.shape.
    It is not valid to request the shape of a hlfir.expr which has no shape.
    """

    name = "hlfir.shape_of"
    expr = operand_def()
    result = result_def()


@irdl_op_definition
class GetExtentOp(IRDLOperation):
    """
    Gets an extent value from a fir.shape. The dimension argument uses C style
    indexing and so should be between 0 and 1 less than the rank of the shape
    """

    name = "hlfir.get_extent"
    shape = operand_def()
    dim = prop_def(IntAttr)
    result = result_def()


@irdl_op_definition
class RegionAssignOp(IRDLOperation):
    """
    This operation can represent Forall and Where assignment when inside an
    hlfir.forall or hlfir.where "ordered assignment tree". It can
    also represent user defined assignments and assignment to vector
    subscripted entities without requiring the materialization of the
    right-hand side temporary copy that may be needed to implement Fortran
    assignment semantic.

    The right-hand side and left-hand side evaluations are held in their
    own regions terminated with hlfir.yield operations (or hlfir.elemental_addr
    for a left-hand side with vector subscript).

    An optional region may be added to implement user defined assignment.
    This region provides two block arguments with the same type as the
    yielded rhs and lhs entities (in that order), or the element type if this
    is an elemental user defined assignment.

    If this optional region is not provided, intrinsic assignment is performed.

    Example: "X = Y",  where "=" is a user defined elemental assignment "foo"
    taking Y by value.

    hlfir.region_assign {
      hlfir.yield %y : !fir.box<!fir.array<?x!f32>>
    } to {
      hlfir.yield %x : !fir.box<!fir.array<?x!fir.type<t>>>
    } user_defined_assignment (%rhs_elt: !fir.ref<f32>) to (%lhs_elt: !fir.ref<!fir.type<t>>) {
      %0 = fir.load %rhs_elt : !fir.ref<f32>
      fir.call @foo(%lhs_elt, %0) : (!fir.ref<!fir.type<t>>, f32) -> ()
    }

    TODO: add optional "realloc" semantics like for hlfir.assign.
    """

    name = "hlfir.region_assign"
    rhs_region = region_def()
    lhs_region = region_def()
    user_defined_assignment = region_def()


@irdl_op_definition
class RegionYieldOp(IRDLOperation):
    """
    Terminator operation that yields an HLFIR value or variable that was computed in
    a region and hold the yielded entity cleanup, if any, into its own region.
    This allows representing any Fortran expression evaluation in its own region so
    that the evaluation can easily be scheduled/moved around in a pass.

    Example: "foo(x)" where foo returns an allocatable array.
    {
      // In some region.
      %0 = fir.call @foo(x) (!fir.ref<f32>) -> !fir.box<fir.heap<!fir.array<?xf32>>>
      hlfir.yield %0 : !fir.box<!fir.heap<!fir.array<?xf32>>> cleanup {
        %1 = fir.box_addr %0 : !fir.box<!fir.heap<!fir.array<?xf32>>> -> !fir.heap<!fir.array<?xf32>>
        %fir.freemem %1 : !fir.heap<!fir.array<?xf32>>
      }
    }
    """

    name = "hlfir.yield"
    entity = operand_def()
    cleanup = region_def()


@irdl_op_definition
class ElementalAddrOp(IRDLOperation):
    """
    Special terminator node for the left-hand side region of an hlfir.region_assign
    to a vector subscripted entity.

    It represents how the address of an element of such entity is computed given
    one based indices.

    It is very similar to hlfir.elemental, except that it does not produce an SSA
    value because there is no hlfir type to describe a vector subscripted entity
    (the codegen of such type would be problematic). Hence, it is tightly linked
    to an hlfir.region_assign by its terminator property.

    An optional cleanup region may be provided if any of the subscript expressions
    of the designator require a cleanup.
    This allows documenting cleanups that cannot be generated after the vector
    subscripted designator usage (that has not been materizaled yet). The cleanups
    will be evaluated after the assignment once the related
    hlfir.region_assign is lowered.

    Example: "X(VECTOR) = Y"

    hlfir.region_assign {
      hlfir.yield %y : !fir.ref<!fir.array<20xf32>>
    } to {
      hlfir.elemental_addr %vector_shape  : !fir.shape<1> {
        ^bb0(%i: index):
        %0 = hlfir.designate %vector (%i)  : (!fir.ref<!fir.array<20xi32>>, index) -> !fir.ref<i32>
        %1 = fir.load %0 : !fir.ref<i32>
        %x_element_addr = hlfir.designate %x (%1)  : (!fir.ref<!fir.array<100xf32>>, i32) -> !fir.ref<f32>
        hlfir.yield %x_element_addr : !fir.ref<f32>
      }
    }
    """

    name = "hlfir.elemental_addr"
    shape = operand_def()
    typeparams = var_operand_def()
    unordered = opt_prop_def(UnitAttr)
    body = region_def()
    cleanup = region_def()

    irdl_options = [AttrSizedOperandSegments(as_property=True)]


@irdl_op_definition
class ForallOp(IRDLOperation):
    """
    This operation allows representing Fortran forall. It computes
    a set of "index-name" values based on lower bound, upper bound,
    and step values whose evaluations are represented in their own
    regions.

    Operations nested in its body region are evaluated in order.
    As opposed to a regular loop, each nested operation is
    fully evaluated for all the values in the "active set of
    index-name" before the next nested operation. In practice, the
    nested operation evaluation may be fused if it is proven that
    they do not have data dependency.

    The "index-name" value is represented as the argument of the
    body region.

    The lower, upper, and step region (if provided), must be terminated
    by hlfir.yield that yields scalar integers.

    The body region must only contain other OrderedAssignmentTreeOpInterface
    operations (like hlfir.region_assign, or other hlfir.forall).

    A Fortran forall with several indices is represented as a nest
    of hlfir.forall.

    All the regions contained in the hlfir.forall must only contain
    code that is pure from a Fortran point of view, except for the
    assignment effect of the hlfir.region_assign.
    This matches Fortran constraint C1037, but requires the outer
    controls to be evaluated outside of the hlfir.forall (these
    controls may have side effects as per Fortran 2018 10.1.4 section).

    Example: FORALL(I=1:10) X(I) = FOO(I)
    hlfir.forall lb {
      hlfir.yield %c1 : index
    } ub {
      hlfir.yield %c10 : index
    } (%i : index) {
      hlfir.region_assign {
        %res = fir.call @foo(%i) : (index) -> f32
        hlfir.yield %res : f32
      } to {
        %xi = hlfir.designate %x(%i) : (!fir.box<!fir.array<?xf32>>, index) -> !fir.ref<f32>
        hlfir.yield %xi : !fir.ref<f32>
      }
    }
    """

    name = "hlfir.forall"
    lb_region = region_def()
    ub_region = region_def()
    step_region = region_def()
    body = region_def()


@irdl_op_definition
class ForallMaskOp(IRDLOperation):
    """
    Fortran Forall can have a scalar mask expression that depends on the
    Forall index-name value.
    hlfir.forall_mask allows representing this mask. The expression
    evaluation is held in the mask region that must yield an i1 scalar
    value.
    An hlfir.forall_mask must be directly nested in the body region of
    an hlfir.forall. It is a separate operation so that it can use the
    index SSA value defined by the hlfir.forall body region.

    Example: "FORALL(I=1:10, SOME_CONDITION(I)) X(I) = FOO(I)"
    hlfir.forall lb {
      hlfir.yield %c1 : index
    } ub {
      hlfir.yield %c10 : index
    } (%i : index) {
      hlfir.forall_mask {
        %mask = fir.call @some_condition(%i) : (index) -> i1
        hlfir.yield %mask : i1
      } do {
        hlfir.region_assign {
          %res = fir.call @foo(%i) : (index) -> f32
          hlfir.yield %res : f32
        } to {
          %xi = hlfir.designate %x(%i) : (!fir.box<!fir.array<?xf32>>, index) -> !fir.ref<f32>
          hlfir.yield %xi : !fir.ref<f32>
        }
      }
    }
    """

    name = "hlfir.forall_mask"
    mask_region = region_def()
    body = region_def()


@irdl_op_definition
class AssignmentMaskOp(IRDLOperation):
    """
    Represent Fortran "where" construct or statement. The mask
    expression evaluation is held in the mask region that must yield
    logical array that has the same shape as all the nested
    hlfir.region_assign left-hand sides, and all the nested hlfir.where
    or hlfir.elsewhere masks.

    The values of the where and elsewhere masks form a control mask that
    controls all the nested hlfir.region_assign: only the array element for
    which the related control mask value is true are assigned. Any right-hand
    side elemental expression is only evaluated for elements where the control
    mask is true. See Fortran standard 2018 section 10.2.3 for more detailed
    about the control mask semantic.

    An hlfir.where must not contain any hlfir.forall but it may be contained
    in such operation. This matches Fortran rules.
    """

    name = "hlfir.where"
    mask_region = region_def()
    body = region_def()


@irdl_op_definition
class ElseWhereOp(IRDLOperation):
    """
    Represent Fortran "elsewhere" construct or statement.

    It has an optional mask region to hold the evaluation of Fortran
    optional elsewhere mask expressions. If this region is provided,
    it must satisfy the same constraints as hlfir.where mask region.

    An hlfir.elsewhere must be the last operation of an hlfir.where or,
    hlfir.elsewhere body, which is enforced by its terminator property.

    Like in Fortran, an hlfir.elsewhere negate the current control mask,
    and if provided, adds the mask the resulting control mask (with a logical
    AND).
    """

    name = "hlfir.elsewhere"
    mask_region = region_def()
    body = region_def()


@irdl_op_definition
class ForallIndexOp(IRDLOperation):
    """
    This operation allows placing an hlfir.forall index in memory with
    the related Fortran index-value name and type.

    So far, lowering needs to manipulate symbols as memory entities.
    This operation allows fulfilling this requirements without allowing
    bare alloca/declare/store inside the body of hlfir.forall, which would
    make their analysis more complex.

    Given Forall index-value cannot be modified it also allows defining
    a canonicalization of all its loads into a fir.convert of the
    hlfir.forall index, which helps simplifying the data dependency analysis
    of hlfir.forall.
    """

    name = "hlfir.forall_index"
    index = operand_def()
    indexname = prop_def(StringAttr)
    result = result_def()


@irdl_op_definition
class CharExtremumOp(IRDLOperation):
    """
    Find the lexicographical minimum or maximum of two or more character
    strings of the same character kind and return the string with the lexicographical
    minimum or maximum number of characters. Example:

    %0 = hlfir.char_extremum min, %arg0, %arg1 : (!fir.ref<!fir.char<1,10>>, !fir.ref<!fir.char<1,20>>) -> !hlfir.expr<!fir.char<1,10>>
    """  # noqa E501

    name = "hlfir.char_extremum"
    predicate = operand_def()
    strings = var_operand_def()
    result = result_def()

    irdl_options = [AttrSizedOperandSegments(as_property=True)]


HLFIR = Dialect(
    "hlfir",
    [
        DeclareOp,
        DesignateOp,
        AssignOp,
        ParentComponentOp,
        ConcatOp,
        AllOp,
        AnyOp,
        CountOp,
        MaxvalOp,
        MinvalOp,
        ProductOp,
        SetLengthOp,
        GetLengthOp,
        SumOp,
        DotProductOp,
        MatmulOp,
        TransposeOp,
        MatmulTransposeOp,
        AssociateOp,
        EndAssociateOp,
        AsExprOp,
        NoReassocOp,
        ElementalOp,
        YieldElementOp,
        ApplyOp,
        NullOp,
        DestroyOp,
        CopyInOp,
        CopyOutOp,
        ShapeOfOp,
        GetExtentOp,
        RegionAssignOp,
        RegionYieldOp,
        ElementalAddrOp,
        ForallOp,
        ForallMaskOp,
        AssignmentMaskOp,
        ElseWhereOp,
        ForallIndexOp,
        CharExtremumOp,
    ],
    [
        ExprType,
    ],
)



import ast
from contextlib import AbstractContextManager
from dataclasses import dataclass
from inspect import getsource
from sys import _getframe  # pyright: ignore[reportPrivateUsage]
from typing import Any

from xdsl.frontend.pyast.program import FrontendProgram
from xdsl.frontend.pyast.python_code_check import PythonCodeCheck


@dataclass
class CodeContext(AbstractContextManager[Any]):
    """
    The CodeContext with block marks the scope in which the code in the custom
    DSL can be written. This code will be translated to xDSL/MLIR.
    """

    program: FrontendProgram
    """
    Underlying frontend program which can be compiled and translated to
    xDSL/MLIR.
    """

    def __enter__(self) -> None:
        # First, get the Python AST from the code.
        frame = _getframe(1)
        self.program.file = frame.f_code.co_filename
        src = getsource(frame)
        python_ast = ast.parse(src)

        # Get all the global information and record it as well. In particular,
        # this contains all the imports.
        self.program.globals = frame.f_globals

        # Find where the program starts.
        for node in ast.walk(python_ast):
            if (
                isinstance(node, ast.With)
                and node.lineno == frame.f_lineno - frame.f_code.co_firstlineno + 1
            ):
                # Found the program AST. Store it for later compilation or
                # execution.
                self.program.stmts = node.body

    def __exit__(self, *args: Any):
        # Having proccessed all the code in the context, check it is well-formed
        # and can be compiled/executed. Additionally, record it for subsequent code generation.
        assert self.program.stmts is not None
        self.program.functions_and_blocks = PythonCodeCheck.run(
            self.program.stmts, self.program.file
        )



# RUN: python %s | filecheck %s

from xdsl.frontend.pyast.context import CodeContext
from xdsl.frontend.pyast.dialects.builtin import f16, f32, f64, i1, i32, i64
from xdsl.frontend.pyast.exception import CodeGenerationException
from xdsl.frontend.pyast.program import FrontendProgram

p = FrontendProgram()
with CodeContext(p):
    # CHECK: arith.addi %{{.*}}, %{{.*}} : i32
    def test_addi_overload(a: i32, b: i32) -> i32:
        return a + b

    # CHECK: arith.andi %{{.*}}, %{{.*}} : i64
    def test_andi_overload(a: i64, b: i64) -> i64:
        return a & b

    # CHECK: arith.muli %{{.*}}, %{{.*}} : i32
    def test_muli_overload(a: i32, b: i32) -> i32:
        return a * b

    # CHECK: arith.subi %{{.*}}, %{{.*}} : i64
    def test_subi_overload(a: i64, b: i64) -> i64:
        return a - b

    # CHECK: arith.shli %{{.*}}, %{{.*}} : i32
    def test_shli_overload(a: i32, b: i32) -> i32:
        return a << b

    # CHECK: arith.shrsi %{{.*}}, %{{.*}} : i64
    def test_shrsi_overload(a: i64, b: i64) -> i64:
        return a >> b

    # CHECK:  arith.cmpi eq, %{{.*}}, %{{.*}} : i32
    def test_cmpi_eq_overload(a: i32, b: i32) -> i1:
        return a == b

    # CHECK: arith.cmpi sle, %{{.*}}, %{{.*}} : i64
    def test_cmpi_le_overload(a: i64, b: i64) -> i1:
        return a <= b

    # CHECK: arith.cmpi slt, %{{.*}}, %{{.*}} : i32
    def test_cmpi_lt_overload(a: i32, b: i32) -> i1:
        return a < b

    # CHECK: arith.cmpi sge, %{{.*}}, %{{.*}} : i64
    def test_cmpi_ge_overload(a: i64, b: i64) -> i1:
        return a >= b

    # CHECK: arith.cmpi sgt, %{{.*}}, %{{.*}} : i32
    def test_cmpi_gt_overload(a: i32, b: i32) -> i1:
        return a > b

    # CHECK: arith.cmpi ne, %{{.*}}, %{{.*}} : i64
    def test_cmpi_ne_overload(a: i64, b: i64) -> i1:
        return a != b

    # CHECK: arith.addf %{{.*}}, %{{.*}} : f16
    def test_addf_overload_f16(a: f16, b: f16) -> f16:
        return a + b

    # CHECK: arith.addf %{{.*}}, %{{.*}} : f32
    def test_addf_overload_f32(a: f32, b: f32) -> f32:
        return a + b

    # CHECK: arith.addf %{{.*}}, %{{.*}} : f64
    def test_addf_overload_f64(a: f64, b: f64) -> f64:
        return a + b

    # CHECK: arith.subf %{{.*}}, %{{.*}} : f16
    def test_subf_overload_f16(a: f16, b: f16) -> f16:
        return a - b

    # CHECK: arith.subf %{{.*}}, %{{.*}} : f32
    def test_subf_overload_f32(a: f32, b: f32) -> f32:
        return a - b

    # CHECK: arith.subf %{{.*}}, %{{.*}} : f64
    def test_subf_overload_f64(a: f64, b: f64) -> f64:
        return a - b

    # CHECK: arith.mulf %{{.*}}, %{{.*}} : f16
    def test_mulf_overload_f16(a: f16, b: f16) -> f16:
        return a * b

    # CHECK: arith.mulf %{{.*}}, %{{.*}} : f32
    def test_mulf_overload_f32(a: f32, b: f32) -> f32:
        return a * b

    # CHECK: arith.mulf %{{.*}}, %{{.*}} : f64
    def test_mulf_overload_f64(a: f64, b: f64) -> f64:
        return a * b


p.compile(desymref=False)
print(p.textual_format())

try:
    with CodeContext(p):
        # CHECK: Binary operation 'FloorDiv' is not supported by type '_Float64' which does not overload '__floordiv__'.
        def test_missing_floordiv_overload_f64(a: f64, b: f64) -> f64:
            # We expect the type error here, since the function doesn't exist on f64
            return a // b

    p.compile(desymref=False)
    print(p.textual_format())
except CodeGenerationException as e:
    print(e.msg)

try:
    with CodeContext(p):
        # CHECK: Comparison operation 'In' is not supported by type '_Float64' which does not overload '__contains__'.
        def test_missing_contains_overload_f64(a: f64, b: f64) -> f64:
            # We expect the type error here, since the function doesn't exist on f64
            return a in b

    p.compile(desymref=False)
    print(p.textual_format())
except CodeGenerationException as e:
    print(e.msg)



from xdsl.dialects import arith, builtin, func, llvm, mpi
from xdsl.dialects.builtin import i32, i64
from xdsl.ir import Attribute, Operation, OpResult
from xdsl.irdl import (
    IRDLOperation,
    Operand,
    irdl_op_definition,
    var_result_def,
)
from xdsl.pattern_rewriter import PatternRewriteWalker
from xdsl.transforms import lower_mpi

info = lower_mpi.MpiLibraryInfo()


def extract_func_call(ops: list[Operation], name: str = "MPI_") -> func.CallOp | None:
    for op in ops:
        if isinstance(op, func.CallOp) and op.callee.string_value().startswith(name):
            return op


def check_emitted_function_signature(
    ops: list[Operation],
    name: str,
    types: tuple[type[Attribute] | None, ...],
):
    call = extract_func_call(ops, name)
    assert call is not None, f"Missing {func.CallOp.name} op to {name} in output!"
    assert len(call.arguments) == len(types)
    for arg, arg_type in zip(call.arguments, types):
        # check that the argument type is correct (if constraint present)
        if arg_type is not None:
            assert isinstance(arg.type, arg_type), (
                f"Expected argument to be of type {arg_type} (got {arg.type} instead)"
            )


@irdl_op_definition
class CreateTestValsOp(IRDLOperation):
    name = "testing.test"
    result = var_result_def()

    @staticmethod
    def get(*types: Attribute):
        return CreateTestValsOp.build(result_types=[list(types)])


def test_lower_mpi_init():
    ops, result = lower_mpi.LowerMpiInit(info).lower(mpi.InitOp.build())

    assert len(result) == 0
    assert len(ops) == 2

    nullop, call = ops

    assert isinstance(call, func.CallOp)
    assert isinstance(nullop, llvm.ZeroOp)
    assert call.callee.string_value() == "MPI_Init"
    assert len(call.arguments) == 2
    assert all(arg == nullop.res for arg in call.arguments)


def test_lower_mpi_finalize():
    ops, result = lower_mpi.LowerMpiFinalize(info).lower(mpi.FinalizeOp.build())

    assert len(result) == 0
    assert len(ops) == 1

    (call,) = ops

    assert isinstance(call, func.CallOp)
    assert call.callee.string_value() == "MPI_Finalize"
    assert len(call.arguments) == 0


def test_lower_mpi_wait_no_status():
    (request,) = CreateTestValsOp.get(mpi.RequestType()).results

    ops, result = lower_mpi.LowerMpiWait(info).lower(mpi.WaitOp(request))

    assert len(result) == 0
    call = extract_func_call(ops)
    assert call is not None
    assert call.callee.string_value() == "MPI_Wait"
    assert len(call.arguments) == 2


def test_lower_mpi_wait_with_status():
    (request,) = CreateTestValsOp.get(mpi.RequestType()).results

    ops, result = lower_mpi.LowerMpiWait(info).lower(
        mpi.WaitOp(request, ignore_status=False)
    )

    assert len(result) == 1
    assert result[0] is not None
    assert isinstance(result[0].type, llvm.LLVMPointerType)
    call = extract_func_call(ops)
    assert call is not None
    assert call.callee.string_value() == "MPI_Wait"
    assert len(call.arguments) == 2
    assert isinstance(call.arguments[1], OpResult)
    assert isinstance(call.arguments[1].owner, llvm.AllocaOp)


def test_lower_mpi_comm_rank():
    ops, result = lower_mpi.LowerMpiCommRank(info).lower(mpi.CommRankOp())

    assert len(result) == 1
    assert result[0] is not None
    assert result[0].type == i32

    # check signature of emitted function call
    # int MPI_Comm_rank(MPI_Comm comm, int *rank)
    check_emitted_function_signature(
        ops,
        "MPI_Comm_rank",
        (None, llvm.LLVMPointerType),
    )


def test_lower_mpi_comm_size():
    ops, result = lower_mpi.LowerMpiCommSize(info).lower(mpi.CommSizeOp())

    assert len(result) == 1
    assert result[0] is not None
    assert result[0].type == i32

    # check signature of emitted function call
    # int MPI_Comm_size(MPI_Comm comm, int *size)
    check_emitted_function_signature(
        ops,
        "MPI_Comm_size",
        (None, llvm.LLVMPointerType),
    )


def test_lower_mpi_send():
    buff, size, dtype, dest, tag = CreateTestValsOp.get(
        llvm.LLVMPointerType.opaque(), i32, mpi.DataType(), i32, i32
    ).results

    ops, result = lower_mpi.LowerMpiSend(info).lower(
        mpi.SendOp(buff, size, dtype, dest, tag)
    )
    """
    Check for function with signature like:
    int MPI_Send(const void *buf, int count, MPI_Datatype datatype, int dest,
             int tag, MPI_Comm comm)
    """

    # send has no return
    assert len(result) == 0

    check_emitted_function_signature(
        ops,
        "MPI_Send",
        (llvm.LLVMPointerType, type(i32), None, type(i32), type(i32), None),
    )


def test_lower_mpi_isend():
    ptr, count, dtype, dest, tag, req = CreateTestValsOp.get(
        llvm.LLVMPointerType.opaque(), i32, mpi.DataType(), i32, i32, mpi.RequestType()
    ).results

    ops, result = lower_mpi.LowerMpiIsend(info).lower(
        mpi.IsendOp(ptr, count, dtype, dest, tag, req)
    )
    """
    Check for function with signature like:
    int MPI_Isend(const void *buf, int count, MPI_Datatype datatype, int dest,
              int tag, MPI_Comm comm, MPI_Request *request)
    """

    # send has no return
    assert len(result) == 0

    check_emitted_function_signature(
        ops,
        "MPI_Isend",
        (
            llvm.LLVMPointerType,
            type(i32),
            None,
            type(i32),
            type(i32),
            None,
            mpi.RequestType,
        ),
    )


def test_lower_mpi_recv_no_status():
    buff, count, dtype, source, tag = CreateTestValsOp.get(
        llvm.LLVMPointerType.opaque(), i32, mpi.DataType(), i32, i32
    ).results
    """
    int MPI_Recv(void *buf, int count, MPI_Datatype datatype, int source, int tag,
             MPI_Comm comm, MPI_Status *status)
    """

    ops, result = lower_mpi.LowerMpiRecv(info).lower(
        mpi.RecvOp(buff, count, dtype, source, tag, ignore_status=True)
    )

    assert len(result) == 0

    check_emitted_function_signature(
        ops,
        "MPI_Recv",
        (
            llvm.LLVMPointerType,
            type(i32),
            None,
            type(i32),
            type(i32),
            None,
            llvm.LLVMPointerType,
        ),
    )


def test_lower_mpi_recv_with_status():
    buff, count, dtype, source, tag = CreateTestValsOp.get(
        llvm.LLVMPointerType.opaque(), i32, mpi.DataType(), i32, i32
    ).results
    """
    int MPI_Recv(void *buf, int count, MPI_Datatype datatype, int source, int tag,
             MPI_Comm comm, MPI_Status *status)
    """

    ops, result = lower_mpi.LowerMpiRecv(info).lower(
        mpi.RecvOp(buff, count, dtype, source, tag, ignore_status=False)
    )

    assert len(result) == 1

    check_emitted_function_signature(
        ops,
        "MPI_Recv",
        (
            llvm.LLVMPointerType,
            type(i32),
            None,
            type(i32),
            type(i32),
            None,
            llvm.LLVMPointerType,
        ),
    )


def test_lower_mpi_irecv():
    ptr, count, dtype, source, tag, req = CreateTestValsOp.get(
        llvm.LLVMPointerType.opaque(), i32, mpi.DataType(), i32, i32, mpi.RequestType()
    ).results
    """
    int MPI_Irecv(void *buf, int count, MPI_Datatype datatype,
            int source, int tag, MPI_Comm comm, MPI_Request *request)
    """

    ops, result = lower_mpi.LowerMpiIrecv(info).lower(
        mpi.IrecvOp(ptr, count, dtype, source, tag, req)
    )

    # recv has no results
    assert len(result) == 0

    check_emitted_function_signature(
        ops,
        "MPI_Irecv",
        (
            llvm.LLVMPointerType,
            type(i32),
            None,
            type(i32),
            type(i32),
            None,
            mpi.RequestType,
        ),
    )


def test_lower_mpi_reduce():
    ptr, count, dtype, root = CreateTestValsOp.get(
        llvm.LLVMPointerType.opaque(), i32, mpi.DataType(), i32
    ).results
    """
    int MPI_Reduce(const void *sendbuf, void *recvbuf, int count,
               MPI_Datatype datatype, MPI_Op op, int root, MPI_Comm comm)
    """

    ops, result = lower_mpi.LowerMpiReduce(info).lower(
        mpi.ReduceOp(ptr, ptr, count, dtype, mpi.MpiOp.MPI_SUM, root)
    )

    # reduce has no results
    assert len(result) == 0

    check_emitted_function_signature(
        ops,
        "MPI_Reduce",
        (
            llvm.LLVMPointerType,
            llvm.LLVMPointerType,
            type(i32),
            None,
            None,
            type(i32),
            None,
        ),
    )


def test_lower_mpi_all_reduce_no_send_buffer():
    ptr, count, dtype = CreateTestValsOp.get(
        llvm.LLVMPointerType.opaque(), i32, mpi.DataType()
    ).results
    """
    int MPI_Allreduce(const void *sendbuf, void *recvbuf, int count,
                  MPI_Datatype datatype, MPI_Op op, MPI_Comm comm)
    """

    ops, result = lower_mpi.LowerMpiAllreduce(info).lower(
        mpi.AllreduceOp(None, ptr, count, dtype, mpi.MpiOp.MPI_SUM)
    )

    # allreduce has no results
    assert len(result) == 0

    check_emitted_function_signature(
        ops,
        "MPI_Allreduce",
        (
            type(i64),
            llvm.LLVMPointerType,
            type(i32),
            None,
            None,
            None,
        ),
    )


def test_mpi_waitall():
    _, count, _ = CreateTestValsOp.get(
        llvm.LLVMPointerType.opaque(), i32, mpi.DataType()
    ).results

    dummy = arith.ConstantOp.from_int_and_width(4, 32)
    alloc_request_op = mpi.AllocateTypeOp(mpi.RequestType, dummy)
    req_op: Operand = alloc_request_op.results[0]
    waitall = mpi.WaitallOp(req_op, count)

    assert waitall.operands[0] == req_op
    assert waitall.operands[1] == count
    assert len(waitall.results) == 0
    # TODO: enhance this test with a lowering


def test_lower_mpi_bcast():
    ptr, count, dtype, root = CreateTestValsOp.get(
        llvm.LLVMPointerType.opaque(), i32, mpi.DataType(), i32
    ).results
    """
    int MPI_Bcast(void *buffer, int count, MPI_Datatype datatype, int root,
              MPI_Comm comm)
    """

    ops, result = lower_mpi.LowerMpiBcast(info).lower(
        mpi.BcastOp(ptr, count, dtype, root)
    )

    # bcast has no results
    assert len(result) == 0

    check_emitted_function_signature(
        ops,
        "MPI_Bcast",
        (
            llvm.LLVMPointerType,
            type(i32),
            None,
            type(i32),
            None,
        ),
    )


def test_lower_mpi_allocate():
    (count,) = CreateTestValsOp.get(i32).results
    op = mpi.AllocateTypeOp(mpi.RequestType, count)

    ops, res = lower_mpi.LowerMpiAllocateType(info).lower(op)

    assert len(res) == 1
    assert len(ops) == 1
    assert isinstance(ops[0], llvm.AllocaOp)


def test_lower_mpi_vec_get():
    mod = builtin.ModuleOp(
        [
            count := CreateTestValsOp.get(i32),
            vec := mpi.AllocateTypeOp(mpi.RequestType, count),
            get := mpi.VectorGetOp(vec, count),
        ]
    )
    # we have to apply this rewrite to that the argument type of the `get`
    # becomes an llvm.ptr
    PatternRewriteWalker(lower_mpi.LowerMpiAllocateType(info)).rewrite_module(mod)

    ops, res = lower_mpi.LowerMpiVectorGet(info).lower(get)

    assert len(res) == 1
    assert res[0] is not None
    assert isinstance(res[0].type, llvm.LLVMPointerType)
    assert len(ops) > 0


def test_lower_mpi_gather():
    ptr, count, dtype, root = CreateTestValsOp.get(
        llvm.LLVMPointerType.opaque(), i32, mpi.DataType(), i32
    ).results

    ops, result = lower_mpi.LowerMpiGatherOp(info).lower(
        mpi.GatherOp(ptr, count, dtype, ptr, count, dtype, root)
    )

    # gather has no results
    assert len(result) == 0

    check_emitted_function_signature(
        ops,
        "MPI_Gather",
        (
            llvm.LLVMPointerType,
            type(i32),
            None,
            llvm.LLVMPointerType,
            type(i32),
            None,
            type(i32),
            None,
        ),
    )


def test_mpi_type_conversion():
    """
    Test that each builtin datatype is correctly mapped to an MPI datatype

    """
    info = lower_mpi.MpiLibraryInfo(
        MPI_UNSIGNED_CHAR=1,
        MPI_UNSIGNED_SHORT=2,
        MPI_UNSIGNED=3,
        MPI_UNSIGNED_LONG_LONG=4,
        MPI_CHAR=5,
        MPI_SHORT=6,
        MPI_INT=7,
        MPI_LONG_LONG_INT=8,
    )

    lowering = lower_mpi.LowerMpiRecv(info)

    from xdsl.dialects.builtin import IntegerType, Signedness, f32, f64, i32, i64

    u64 = IntegerType(64, Signedness.UNSIGNED)
    u32 = IntegerType(32, Signedness.UNSIGNED)

    checks = [
        (f32, info.MPI_FLOAT),
        (f64, info.MPI_DOUBLE),
        (i32, info.MPI_INT),
        (u32, info.MPI_UNSIGNED),
        (i64, info.MPI_LONG_LONG_INT),
        (u64, info.MPI_UNSIGNED_LONG_LONG),
    ]

    for width in (8, 16):
        for sign in (Signedness.UNSIGNED, Signedness.SIGNLESS, Signedness.SIGNED):
            sign_str = "UNSIGNED_" if sign == Signedness.UNSIGNED else ""
            name = "CHAR" if width == 8 else "SHORT"
            mpi_type = IntegerType(width, sign)
            checks.append((mpi_type, getattr(info, f"MPI_{sign_str}{name}")))

    for type, target in checks:
        # we test a private member function here, so we ignore pyright
        translate_to_mpi_type = (
            lowering._translate_to_mpi_type  # pyright: ignore[reportPrivateUsage]
        )
        assert translate_to_mpi_type(type) == target
